@@ -1,6 +1,6 @@
 /* cache flushing for the HP-PA architecture
 
-  Copyright (C) 1995,1996,1997,1998 Free Software Foundation, Inc.
+  Copyright (C) 1995,1996,1997,1998,2003 Free Software Foundation, Inc.
 
   This file is part of Gforth.
 
@@ -19,6 +19,20 @@
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111, USA.
 */
 
+#include <stddef.h>
+
+void cacheflush(void * address, size_t size, size_t linewidth)
+{
+  void *p=(void *)((size_t)address & (-linewidth));
+
+  for(; p<address+size;)
+    asm volatile("fdc (%0)\n\t"
+		 "sync\n\t"
+		 "fic,m %1(%0)\n\t"
+		 "sync" : "+r"(p) : "r"(linewidth) : "memory" );
+}
+
+#if 0
 void cacheflush(void * address, int size, int linewidth)
 {
   int i;
@@ -31,3 +45,4 @@ void cacheflush(void * address, int size, int linewidth)
 		 "fic,m %1(%0)\n\t"
 		 "sync" : : "r" (address), "r" (linewidth) : "memory" );
 }
+#endif
