@@ -1,6 +1,6 @@
 /* common header file
 
-  Copyright (C) 1995,1996,1997,1998,2000 Free Software Foundation, Inc.
+  Copyright (C) 1995,1996,1997,1998,2000,2003 Free Software Foundation, Inc.
 
   This file is part of Gforth.
 
@@ -20,6 +20,13 @@
 */
 
 #include "config.h"
+#include <stdio.h>
+#include <sys/time.h>
+#include <unistd.h>
+
+#if defined(DOUBLY_INDIRECT)||defined(INDIRECT_THREADED)||defined(VM_PROFILING)
+#define NO_DYNAMIC
+#endif
 
 #if defined(DOUBLY_INDIRECT)
 #  undef DIRECT_THREADED
@@ -27,6 +34,13 @@
 #  define INDIRECT_THREADED
 #endif
 
+#if defined(GFORTH_DEBUGGING)
+#  undef USE_TOS
+#  undef USE_FTOS
+#  define USE_NO_TOS
+#  define USE_NO_FTOS
+#endif
+
 #include <limits.h>
 
 #if defined(NeXT)
@@ -70,15 +84,13 @@ typedef struct {
   UCell lo;
 } UDCell;
 
-#define FETCH_DCELL(d,lo,hi)	((d)=(typeof(d)){(hi),(lo)})
-#define STORE_DCELL(d,low,high)	({ \
-				     typeof(d) _d = (d); \
-				     (low) = _d.lo; \
-				     (high)= _d.hi; \
-				 })
-
-#define LONG2UD(l)	({UDCell _ud; _ud.hi=0; _ud.lo=(Cell)(l); _ud;})
-#define UD2LONG(ud)	((long)(ud.lo))
+#if SMALL_OFF_T
+#define OFF2UD(o) ({UDCell _ud; _ud.hi=0; _ud.lo=(Cell)(o); _ud;})
+#define UD2OFF(ud) ((ud).lo)
+#else /* !SMALL_OFF_T */
+#define OFF2UD(o) ({UDCell _ud; off_t _o=(o); _ud.hi=_o>>CELL_BITS; _ud.lo=(Cell)_o; _ud;})
+#define UD2OFF(ud) ({UDCell _ud=(ud); (((off_t)_ud.hi)<<CELL_BITS)+_ud.lo;})
+#endif /* !SMALL_OFF_T */
 #define DZERO		((DCell){0,0})
 
 #else /* ! defined(BUGGY_LONG_LONG) */
@@ -87,48 +99,47 @@ typedef struct {
 typedef DOUBLE_CELL_TYPE DCell;
 typedef unsigned DOUBLE_CELL_TYPE UDCell;
 
-typedef union {
-  struct {
-#ifdef WORDS_BIGENDIAN
-    Cell high;
-    UCell low;
-#else
-    UCell low;
-    Cell high;
-#endif;
-  } cells;
-  DCell dcell;
-} Double_Store;
-
-#define FETCH_DCELL(d,lo,hi)	({ \
-				     Double_Store _d; \
-				     _d.cells.low = (lo); \
-				     _d.cells.high = (hi); \
-				     (d) = _d.dcell; \
-				 })
-
-#define STORE_DCELL(d,lo,hi)	({ \
-				     Double_Store _d; \
-				     _d.dcell = (d); \
-				     (lo) = _d.cells.low; \
-				     (hi) = _d.cells.high; \
-				 })
-
-#define LONG2UD(l)	((UDCell)(l))
-#define UD2LONG(ud)	((long)(ud))
+#define OFF2UD(o)	((UDCell)(o))
+#define UD2OFF(ud)	((off_t)(ud))
 #define DZERO		((DCell)0)
 
 #endif /* ! defined(BUGGY_LONG_LONG) */
 
-#ifdef DIRECT_THREADED
-typedef Label Xt;
+typedef union {
+  struct {
+#if defined(WORDS_BIGENDIAN)||defined(BUGGY_LONG_LONG)
+    Cell high;
+    UCell low;
 #else
-typedef Label *Xt;
+    UCell low;
+    Cell high;
 #endif
+  } cells;
+  DCell d;
+  UDCell ud;
+} Double_Store;
 
-#if !defined(DIRECT_THREADED)
-/* i.e. indirect threaded our doubly indirect threaded */
-/* the direct threaded version is machine dependent and resides in machine.h */
+#define FETCH_DCELL_T(d_,lo,hi,t_)	({ \
+				     Double_Store _d; \
+				     _d.cells.low = (lo); \
+				     _d.cells.high = (hi); \
+				     (d_) = _d.t_; \
+				 })
+
+#define STORE_DCELL_T(d_,lo,hi,t_)	({ \
+				     Double_Store _d; \
+				     _d.t_ = (d_); \
+				     (lo) = _d.cells.low; \
+				     (hi) = _d.cells.high; \
+				 })
+
+#define vm_twoCell2d(lo,hi,d_)  FETCH_DCELL_T(d_,lo,hi,d);
+#define vm_twoCell2ud(lo,hi,d_) FETCH_DCELL_T(d_,lo,hi,ud);
+
+#define vm_d2twoCell(d_,lo,hi)  STORE_DCELL_T(d_,lo,hi,d);
+#define vm_ud2twoCell(d_,lo,hi) STORE_DCELL_T(d_,lo,hi,ud);
+
+typedef Label *Xt;
 
 /* PFA gives the parameter field address corresponding to a cfa */
 #define PFA(cfa)	(((Cell *)cfa)+2)
@@ -141,13 +152,11 @@ typedef Label *Xt;
 #if !defined(DOUBLY_INDIRECT)
 #  define DOES_CA (symbols[DODOES])
 #else /* defined(DOUBLY_INDIRECT) */
-#  define DOES_CA ((Label)&symbols[DODOES])
+#  define DOES_CA ((Label)&xts[DODOES])
 #endif /* defined(DOUBLY_INDIRECT) */
 
 
 
-#define DOES_CODE(cfa)	({Xt _cfa=(Xt)(cfa); \
-			  (Xt *)(_cfa[0]==DOES_CA ? _cfa[1] : NULL);})
 #define DOES_CODE1(cfa)	((Xt *)(cfa[1]))
 /* MAKE_CF creates an appropriate code field at the cfa;
    ca is the code address */
@@ -155,16 +164,15 @@ typedef Label *Xt;
 /* make a code field for a defining-word-defined word */
 #define MAKE_DOES_CF(cfa,does_code)  ({MAKE_CF(cfa,DOES_CA);	\
 				       ((Cell *)cfa)[1] = (Cell)(does_code);})
-/* the does handler resides between DOES> and the following Forth code */
-/* not needed in indirect threaded code */
-#if defined(DOUBLY_INDIRECT)
-#define MAKE_DOES_HANDLER(addr)	MAKE_CF(addr, ((Label)&symbols[DOESJUMP]))
-#else /* !defined(DOUBLY_INDIRECT) */
-#define MAKE_DOES_HANDLER(addr)	0
-#endif /* !defined(DOUBLY_INDIRECT) */
-#endif /* !defined(DIRECT_THREADED) */
 
-#ifdef DEBUG
+#ifdef GFORTH_DEBUGGING
+#define NAME(string) { saved_ip=ip; asm(""); }
+/* the asm here is to avoid reordering of following stuff above the
+   assignment; this is an old-style asm (no operands), and therefore
+   is treated like "asm volatile ..."; i.e., it prevents most
+   reorderings across itself.  We want the assignment above first,
+   because the stack loads may already cause a stack underflow. */
+#elif DEBUG
 #	define	NAME(string)	fprintf(stderr,"%08lx: "string"\n",(Cell)ip);
 #else
 #	define	NAME(string)
@@ -177,28 +185,28 @@ typedef Label *Xt;
 #ifndef FLUSH_ICACHE
 #warning flush-icache probably will not work (see manual)
 #	define FLUSH_ICACHE(addr,size)
-#endif
-
-#if defined(DIRECT_THREADED)
-#define CACHE_FLUSH(addr,size) FLUSH_ICACHE(addr,size)
-#else
-#define CACHE_FLUSH(addr,size)
+#warning no FLUSH_ICACHE, turning off dynamic native code by default
+#undef NO_DYNAMIC_DEFAULT
+#define NO_DYNAMIC_DEFAULT 1
 #endif
 
 #ifdef USE_TOS
-#define IF_TOS(x) x
+#define IF_spTOS(x) x
 #else
-#define IF_TOS(x)
-#define TOS (sp[0])
+#define IF_spTOS(x)
+#define spTOS (sp[0])
 #endif
 
 #ifdef USE_FTOS
-#define IF_FTOS(x) x
+#define IF_fpTOS(x) x
 #else
-#define IF_FTOS(x)
-#define FTOS (fp[0])
+#define IF_fpTOS(x)
+#define fpTOS (fp[0])
 #endif
 
+#define IF_rpTOS(x)
+#define rpTOS (rp[0])
+
 typedef struct {
   Address base;		/* base address of image (0 if relocatable) */
   UCell checksum;	/* checksum of ca's to protect against some
@@ -213,7 +221,7 @@ typedef struct {
   Xt *boot_entry;	/* initial ip for booting (in BOOT) */
   Xt *throw_entry;	/* ip after signal (in THROW) */
   Cell unused1;		/* possibly tib stack size */
-  Cell unused2;
+  Label *xt_base;         /* base of DOUBLE_INDIRECT xts[], for comp-i.fs */
   Address data_stack_base; /* this and the following fields are initialized by the loader */
   Address fp_stack_base;
   Address return_stack_base;
@@ -221,9 +229,56 @@ typedef struct {
 } ImageHeader;
 /* the image-header is created in main.fs */
 
+struct Longname {
+  struct Longname *next;  /* the link field for old hands */
+  Cell		countetc;
+  char		name[0];
+};
+
+#define LONGNAME_COUNT(np)	((np)->countetc & (((~((UCell)0))<<3)>>3))
+
+struct Cellpair {
+  Cell n1;
+  Cell n2;
+};
+
+struct Cellquad {
+  Cell n1;
+  Cell n2;
+  Cell n3;
+  Cell n4;
+};
+
+#define IOR(flag)	((flag)? -512-errno : 0)
+
 Label *engine(Xt *ip, Cell *sp, Cell *rp, Float *fp, Address lp);
+Label *engine2(Xt *ip, Cell *sp, Cell *rp, Float *fp, Address lp);
+Label *engine3(Xt *ip, Cell *sp, Cell *rp, Float *fp, Address lp);
+
+/* engine/prim support routines */
 Address my_alloc(Cell size);
+char *cstr(Char *from, UCell size, int clear);
 char *tilde_cstr(Char *from, UCell size, int clear);
+DCell timeval2us(struct timeval *tvp);
+void cmove(Char *c_from, Char *c_to, UCell u);
+void cmove_up(Char *c_from, Char *c_to, UCell u);
+Cell compare(Char *c_addr1, UCell u1, Char *c_addr2, UCell u2);
+struct Longname *listlfind(Char *c_addr, UCell u, struct Longname *longname1);
+struct Longname *hashlfind(Char *c_addr, UCell u, Cell *a_addr);
+struct Longname *tablelfind(Char *c_addr, UCell u, Cell *a_addr);
+UCell hashkey1(Char *c_addr, UCell u, UCell ubits);
+struct Cellpair parse_white(Char *c_addr1, UCell u1);
+Cell rename_file(Char *c_addr1, UCell u1, Char *c_addr2, UCell u2);
+struct Cellquad read_line(Char *c_addr, UCell u1, Cell wfileid);
+struct Cellpair file_status(Char *c_addr, UCell u);
+Cell to_float(Char *c_addr, UCell u, Float *rp);
+Float v_star(Float *f_addr1, Cell nstride1, Float *f_addr2, Cell nstride2, UCell ucount);
+void faxpy(Float ra, Float *f_x, Cell nstridex, Float *f_y, Cell nstridey, UCell ucount);
+
+/* signal handler stuff */
+void install_signal_handlers(void);
+typedef void Sigfunc(int);
+Sigfunc *bsd_signal(int signo, Sigfunc *func);
 
 /* dblsub routines */
 DCell dnegate(DCell d1);
@@ -235,16 +290,53 @@ DCell fmdiv (DCell num, Cell denom);
 
 Cell memcasecmp(const Char *s1, const Char *s2, Cell n);
 
+void vm_print_profile(FILE *file);
+void vm_count_block(Xt *ip);
+
+/* dynamic superinstruction stuff */
+Label compile_prim(Label prim);
+void compile_prim1(Cell *start);
+void finish_code(void);
+int forget_dyncode(Address code);
+Label decompile_code(Label prim);
+
 extern int offset_image;
 extern int die_on_signal;
 extern UCell pagesize;
 extern ImageHeader *gforth_header;
+extern Label *vm_prims;
+extern Label *xts;
+extern Cell npriminfos;
+
+#ifdef HAS_DEBUG
+extern int debug;
+#else
+# define debug 0
+#endif
+
+extern Cell *SP;
+extern Float *FP;
+extern Address UP;
 
 #ifdef GFORTH_DEBUGGING
-extern Xt *ip;
+extern Xt *saved_ip;
 extern Cell *rp;
 #endif
 
+#ifdef NO_IP
+extern Label next_code;
+#endif
+
+#ifdef HAS_FILE
+extern char* fileattr[6];
+extern char* pfileattr[6];
+extern int ufileattr[6];
+#endif
+
+#ifdef PRINT_SUPER_LENGTHS
+Cell prim_length(Cell prim);
+void print_super_lengths();
+#endif
 
 /* declare all the functions that are missing */
 #ifndef HAVE_ATANH
@@ -272,3 +364,4 @@ extern unsigned long int strtoul(const char *nptr, char **endptr, int base);
 #endif
 
 
+#define GROUP(x, n)
